Package: np
Version: 0.60-20
Date: 2026-02-11
Imports: boot, cubature, methods, quadprog, quantreg, stats
Suggests: crs, MASS, logspline, ks, testthat
Title: Nonparametric Kernel Smoothing Methods for Mixed Data Types
Authors@R: c(person(given = "Jeffrey S.", family = "Racine", role = c("aut","cre"), email = "racinej@mcmaster.ca"),
             person(given = "Tristen", family = "Hayfield", role = c("aut"), email = "tristen.hayfield@gmail.com"))
Maintainer: Jeffrey S. Racine <racinej@mcmaster.ca>
Description: Nonparametric (and semiparametric) kernel methods that seamlessly handle a mix of continuous, unordered, and ordered factor data types. We would like to gratefully acknowledge support from the Natural Sciences and Engineering Research Council of Canada (NSERC, <https://www.nserc-crsng.gc.ca/>), the Social Sciences and Humanities Research Council of Canada (SSHRC, <https://www.sshrc-crsh.gc.ca/>), and the Shared Hierarchical Academic Research Computing Network (SHARCNET, <https://sharcnet.ca/>). We would also like to acknowledge the contributions of the GNU GSL authors. In particular, we adapt the GNU GSL B-spline routine gsl_bspline.c adding automated support for quantile knots (in addition to uniform knots), providing missing functionality for derivatives, and for extending the splines beyond their endpoints.
License: GPL
URL: https://github.com/JeffreyRacine/R-Package-np
BugReports: https://github.com/JeffreyRacine/R-Package-np/issues
Repository: CRAN
NeedsCompilation: yes
Packaged: 2026-02-11 21:12:26 UTC; jracine
Author: Jeffrey S. Racine [aut, cre],
  Tristen Hayfield [aut]
Date/Publication: 2026-02-12 06:10:12 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 05:43:45 UTC; unix
Archs: np.so.dSYM
